"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.WorkerQueue = void 0;
const bullmq_1 = require("bullmq");
const events_1 = require("events");
const ioredis_1 = __importDefault(require("ioredis"));
const logger_1 = require("../logger");
const workerQueueId = `lage:${process.env.LAGE_WORKER_QUEUE_ID || "default"}`;
const workerPubSubChannel = `lage_pubsub_${process.env.LAGE_WORKER_QUEUE_ID || "default"}`;
class WorkerQueue {
    constructor(config) {
        this.config = config;
        const redisOptions = config.workerQueueOptions.connection;
        this.coordinatorRedis = new ioredis_1.default(redisOptions);
    }
    async initializeAsWorker(workerQueueProcessor) {
        if (this.mode) {
            throw new Error(`WorkerQueue already initialized as ${this.mode}!`);
        }
        this.mode = "worker";
        const { config } = this;
        // Convert the simpler workerQueueProcessor to a full-fledged bullmq processor
        const processor = workerQueueProcessor;
        this.worker = new bullmq_1.Worker(workerQueueId, processor, {
            connection: config.workerQueueOptions.connection,
            sharedConnection: true,
            concurrency: config.concurrency,
        });
        const pubSubListener = async (channel, message) => {
            if (workerPubSubChannel === channel) {
                if (message === "done") {
                    await this.worker.close(true);
                    await this.worker.disconnect();
                    this.coordinatorRedis.off("message", pubSubListener);
                    await this.coordinatorRedis.unsubscribe(workerPubSubChannel);
                    await this.coordinatorRedis.quit();
                    this.coordinatorRedis.disconnect();
                }
            }
        };
        await this.coordinatorRedis.subscribe(workerPubSubChannel);
        this.coordinatorRedis.on("message", pubSubListener);
        await this.worker.waitUntilReady();
        return this.worker;
    }
    async initializeAsMaster(maxTargets = 500) {
        if (this.mode) {
            throw new Error(`WorkerQueue already initialized as ${this.mode}!`);
        }
        this.mode = "master";
        const { config } = this;
        if (!this.workerQueue) {
            logger_1.logger.verbose("[dist] creating work queue ");
            this.workerQueue = new bullmq_1.Queue(workerQueueId, Object.assign({}, config.workerQueueOptions));
            logger_1.logger.verbose("[dist] clearing work queue");
            await this.workerQueue.drain(false);
            await this.workerQueue.clean(0, 3000);
            await this.workerQueue.waitUntilReady();
            this.queueEvents = new bullmq_1.QueueEvents(workerQueueId);
            this.queueEvents.setMaxListeners(maxTargets);
            logger_1.logger.verbose("[dist] work queue ready");
        }
        return this.workerQueue;
    }
    async addJob(id, jobData) {
        if (this.workerQueue && this.queueEvents) {
            const job = await this.workerQueue.add(id, jobData !== null && jobData !== void 0 ? jobData : {});
            const emitter = new events_1.EventEmitter();
            const completeHandler = async ({ jobId, returnvalue }) => {
                if (job.id === jobId) {
                    this.queueEvents.off("completed", completeHandler);
                    this.queueEvents.off("failed", failedHandler);
                    emitter.emit("completed", { jobId, returnvalue });
                }
            };
            const failedHandler = ({ jobId, failedReason }) => {
                if (job.id === jobId) {
                    this.queueEvents.off("completed", completeHandler);
                    this.queueEvents.off("failed", failedHandler);
                    emitter.emit("failed", { jobId, failedReason });
                }
            };
            this.queueEvents.on("completed", completeHandler);
            this.queueEvents.on("failed", failedHandler);
            return emitter;
        }
    }
    async closing() {
        var _a;
        return await ((_a = this.workerQueue) === null || _a === void 0 ? void 0 : _a.closing);
    }
    async close() {
        if (this.coordinatorRedis) {
            await this.coordinatorRedis.publish(workerPubSubChannel, "done");
            await this.coordinatorRedis.quit();
            this.coordinatorRedis.disconnect();
        }
        if (this.workerQueue) {
            await this.workerQueue.drain(false);
            await this.workerQueue.close();
            await this.workerQueue.disconnect();
        }
        if (this.queueEvents) {
            await this.queueEvents.close();
        }
        logger_1.logger.verbose("[dist] closed work queue");
    }
}
exports.WorkerQueue = WorkerQueue;
//# sourceMappingURL=WorkerQueue.js.map